#!/usr/bin/env node

import * as fs from 'fs';
import * as path from 'path';

import * as commandLineUsage from 'command-line-usage';

import * as exampleConfig from '../config.example.json';
import { getLogger, setLogLevels } from './loggers';
import QQBot from './koishi';
import Worker from './twitter';
import { status, sub, unsub } from './command';

const logger = getLogger();

const sections: commandLineUsage.Section[] = [
  {
    header: 'GoCQHTTP IPWIKI Bot',
    content: 'The QQ Bot that updates Idoly Pride BWIKI.',
  },
  {
    header: 'Synopsis',
    content: [
      '$ cq-ipwiki-bot {underline config.json}',
      '$ cq-ipwiki-bot {bold --help}',
    ],
  },
  {
    header: 'Documentation',
    content: [
      'Project home: {underline https://github.com/CL-Jeremy/mirai-twitter-bot}',
      'Example config: {underline https://git.io/JJ0jN}',
    ],
  },
];

const usage = commandLineUsage(sections);

const args = process.argv.slice(2);

if (args.length === 0 || args[0] === 'help' || args[0] === '-h' || args[0] === '--help') {
  console.log(usage);
  process.exit(0);
}

const configPath = args[0];

type Config = typeof exampleConfig;
let config: Config;
try {
  config = JSON.parse(fs.readFileSync(path.resolve(configPath), 'utf8')) as Config;
} catch (e) {
  console.log('Failed to parse config file: ', configPath);
  console.log(usage);
  process.exit(1);
}

const requiredFields = [
  'twitter_consumer_key', 'twitter_consumer_secret', 'twitter_access_token_key', 'twitter_access_token_secret',
  'cq_bot_qq', 'bilibili_cookie_sessdata'
];

const warningFields = [
  'cq_ws_host', 'cq_ws_port', 'cq_access_token',
];

const optionalFields = [
  'lockfile', 'work_interval', 'loglevel',
].concat(warningFields);

if (requiredFields.some((value) => config[value] === undefined)) {
  console.log(`${requiredFields.join(', ')} are required`);
  process.exit(1);
}

optionalFields.forEach(key => {
  if (config[key] === undefined || typeof(config[key]) !== typeof (exampleConfig[key])) {
    if (warningFields.includes(key)) logger.warn(`${key} is undefined, use ${exampleConfig[key] || 'empty string'} as default`);
    config[key] = exampleConfig[key as keyof Config];
  }
});

setLogLevels(config.loglevel);

let lock: ILock;
if (fs.existsSync(path.resolve(config.lockfile))) {
  try {
    lock = JSON.parse(fs.readFileSync(path.resolve(config.lockfile), 'utf8')) as ILock;
  } catch (err) {
    logger.error(`Failed to parse lockfile ${config.lockfile}: `, err);
    lock = {
      offset: '0',
      lastActions: [],
      subscribers: [],
      updatedAt: undefined,
    };
  }
  fs.access(path.resolve(config.lockfile), fs.constants.W_OK, err => {
    if (err) {
      logger.fatal(`cannot write lockfile ${path.resolve(config.lockfile)}, permission denied`);
      process.exit(1);
    }
  });
} else {
  lock = {
    offset: '0',
    lastActions: [],
    subscribers: [],
    updatedAt: undefined,
  };
  try {
    fs.writeFileSync(path.resolve(config.lockfile), JSON.stringify(lock));
  } catch (err) {
    logger.fatal(`cannot write lockfile ${path.resolve(config.lockfile)}, permission denied`);
    process.exit(1);
  }
}

const qq = new QQBot({
  access_token: config.cq_access_token,
  host: config.cq_ws_host,
  port: config.cq_ws_port,
  bot_id: config.cq_bot_qq,
  status: (c, a, cb) => status(c, a, cb, lock),
  sub: (c, a, cb) => sub(c, a, cb, lock, config.lockfile),
  unsub: (c, a, cb) => unsub(c, a, cb, lock, config.lockfile),
});

const worker = new Worker({
  consumerKey: config.twitter_consumer_key,
  consumerSecret: config.twitter_consumer_secret,
  accessTokenKey: config.twitter_access_token_key,
  accessTokenSecret: config.twitter_access_token_secret,
  wikiSessionCookie: config.bilibili_cookie_sessdata,
  lock,
  lockfile: config.lockfile,
  workInterval: config.work_interval,
  bot: qq,
});
worker.launch();

qq.connect();
